#if defined(WIN32)
  #include <windows.h>
#elif defined(unix)
  #include <pthread.h>
#else
  #error Unknown OS!
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>

#include <atl/integer.h>
#include <atl/list.h>
#include <atl/hashmap.h>
#include <atl/stringbuffer.h>
#include <atl/void_ast.h>
#include <ts/system.h>
#include <generator.h>
#include <utils/network.h>
#include "synchronizer.h"

/* Define if needed. If DEBUG_SYNC is defined, but DEBUG_SYNC_HEAVY is not defined,
 StdOut logs of all processes will be of the same size, except for messages about logging,
 if they have enough time to exchange all the data. */
/* #define DEBUG_SYNC */
/* #define DEBUG_SYNC_HEAVY */

#ifdef DEBUG_SYNC_HEAVY
  #undef  DEBUG_SYNC
  #define DEBUG_SYNC
#endif

/*
#include <utils/assertion.h>
#define ASSERT0	assertion
#define ASSERT1	assertion
#define ASSERT2	assertion
*/

static void ASSERT2(bool expr, char *param0, int param1, int param2) {
	if (!expr) {
		printf(param0, param1, param2);
		printf("\n");
		fflush(stdout);
		exit(1);
	}
}
static void ASSERT1(bool expr, char *param0, int param1) {
	ASSERT2(expr, param0, param1, 0);
}
static void ASSERT0(bool expr, char *param0) {
	ASSERT2(expr, param0, 0, 0);
}


/* Performance tuning */
#define SMALL_MESSAGE_SIZE	512
#define DEFAULT_HASH_SIZE	1024
#define EXCHANGE_CHUNK		256
#define EXPECTED_LOG_SIZE	65536


/* Message types */
#define MTYPE_TRANS	1
#define MTYPE_FINISH	10

/* Message sources */
#define SRC_LOAD_LOG	-1


static void *sure_realloc(void *ptr, int newSize)
{
	if (newSize == 0) {	/* In some systems, realloc(p, 0) does not work as free(p) */
		if (ptr != NULL)
			free(ptr);
		return NULL;
	}

	ptr = realloc(ptr, newSize);
	ASSERT0(ptr != NULL, "NetFSM synchronizer: out of memory");

	return ptr;
}

static int hash_TE_ArcId(TE_ArcId *arc)
{
/* The multipliers were chosen to fit scenario0. They may be bad for some other scenario!
 Even number (10) is used to handle 1-to-1 correlation between index and end_node. */

int res = hash_String( r(arc->start_node_id->id) ) * 17
		+ arc->stimulus_id.index * 10
		+ hash_String( r(arc->end_node_id->id) );

	destroy(arc);
	return res;
}

static void activate_shutdown( Synchronizer *sync );
static void send_message_out( Synchronizer *sync, TE_ArcId *arc );
static void do_load_log( Synchronizer *sync );
static void receive_message( Synchronizer *sync, int inChannel, /*no r()*/TE_ArcList *in_updates );

/************************************************/

struct OutInfo {
	char *outHost;
	unsigned short outPort;
};

struct Synchronizer {
	enum Sync_State {
		CREATED, START_CONF, READY,
		THREAD_RUNNING /* Or finished, but not joined yet */,
		FINALIZED
	} state;

	TE_Graph *graph;
	TestEngine *te;

	int in_socket;
	int num_in;
	int *in_channels;
	int *read_poll;
	bool *shutdown_received;

	struct OutInfo *out_infos;
	int num_out;
	int *out_channels;

	jmp_buf io_exception;
	bool *in_failed, *out_failed;
	volatile int in_fail_count, out_fail_count;

	FILE *save_log;
	FILE *load_log;
	bool load_then_save, simu_load;
	int skip_save_log;

	/* set B with counters */
	HashMap *receive_counter;
	volatile int receive_counter_size;

	HashMap *previous_to_local;

	TSSemaphore write_sema;
	List *write_queue;
	volatile int write_queue_size, writer_in_processing;

#if defined(WIN32)
	HANDLE reader_thread, writer_thread;
#elif defined(unix)
	pthread_t reader_thread, writer_thread;
#endif

	int thread_exchange_delay;
	volatile bool thread_death_required, shutdown_activated;

	bool tolerate_neighbour_deaths;

	/* stats */
	volatile int completed_transactions;
	volatile int loaded_arcs_count, saved_arcs_count;
};

Synchronizer *create_Synchronizer( TE_Graph *g, TestEngine *t )
{
Synchronizer *sync = sure_realloc( NULL, sizeof(struct Synchronizer) );

	memset( sync, 0, sizeof(struct Synchronizer) );

	sync->graph = g;
	sync->te = t;

	sync->thread_exchange_delay = 1;

	sync->state = CREATED;
	return sync;
}

void destroy_Synchronizer( Synchronizer *sync )
{
	switch (sync->state) {
	/* case CREATED:	break; */
	case START_CONF:
		close_socket( sync->in_socket );
		break;
	case READY:
		sync_Shutdown( sync );
		break;
	case THREAD_RUNNING:
		sync_FinishThread( sync );
		break;
	/* case FINALIZED:	break; */
	}

	free(sync);
}

/************************************************/
/*                 Configuring                  */
/************************************************/

int sync_SetInputConf( Synchronizer *sync, unsigned short inPort, int numInConnections )
{
	if (sync->state != CREATED) {
		printf("sync_SetInputConf: illegal state.\n");
		return -1;
	}
	if (numInConnections <= 0) {
		printf("sync_SetInputConf: invalid number of in connections.\n");
		return -1;
	}

	sync->in_socket = create_server(inPort);
	if (sync->in_socket < 0)
		return -1;

	sync->num_in = numInConnections;
#ifdef DEBUG_SYNC
	printf("Listening TCP socket on port %i, awaiting %i connections.\n",
		(int)inPort, numInConnections);
#endif

	sync->state = START_CONF;
	return 0;
}

int sync_AddOutConnection( Synchronizer *sync, char *outHost, unsigned short outPort )
{
int l;
int idx = sync->num_out;

	if (sync->state != START_CONF) {
		printf("sync_AddOutConnection: illegal state.\n");
		return -1;
	}

	sync->out_infos = sure_realloc( sync->out_infos,
		(idx+1) * sizeof(struct OutInfo) );
	sync->out_infos[ sync->num_out ].outPort = outPort;

	l = strlen(outHost) + 1;
	sync->out_infos[ idx ].outHost = sure_realloc( NULL, l );
	memcpy( sync->out_infos[ idx ].outHost, outHost, l );

	sync->num_out++;
	return 0;
}

int sync_TolerateNeighbourDeaths( Synchronizer *sync, bool tolerate )
{
	if (sync->state != START_CONF) {
		printf("sync_TolerateNeighbourDeaths: illegal state.\n");
		return -1;
	}

	sync->tolerate_neighbour_deaths = tolerate;

	return 0;
}

int sync_SaveLog( Synchronizer *sync, const char *fileName )
{
	if (sync->state != START_CONF && sync->state != CREATED) {
		printf("sync_SaveLog: illegal state.\n");
		return -1;
	}
	if (sync->save_log != NULL) {
		printf("sync_SaveLog: log already set.\n");
		return -1;
	}
	if (sync->load_log != NULL && sync->load_then_save) {
		printf("sync_SaveLog: log to append already set.\n");
		return -1;
	}

	sync->save_log = fopen( fileName, "wb+" );
	if (sync->save_log == NULL) {
		perror("sync_SaveLog: fopen");
		return -1;
	}

#ifdef DEBUG_SYNC
	printf("Writing log to '%s'\n", fileName);
#endif
	return 0;
}

int sync_LoadLog( Synchronizer *sync, const char *fileName, bool thenAppend, bool simLoad )
{
	if (sync->state != START_CONF && sync->state != CREATED) {
		printf("sync_LoadLog: illegal state.\n");
		return -1;
	}
	if (sync->load_log != NULL) {
		printf("sync_LoadLog: log to load already set.\n");
		return -1;
	}

	if (thenAppend) {
		if (sync->save_log != NULL) {
			printf("sync_LoadLog: log to save already set.\n");
			return -1;
		}

		sync->load_log = fopen( fileName, "rb+" );
		if (sync->load_log == NULL && sync->save_log == NULL) {
#ifdef DEBUG_SYNC
			printf("Cannot read log from '%s', trying to write...\n", fileName);
#endif
			/* sync->load_then_save and sync->simu_load remain false */
			return sync_SaveLog( sync, fileName );
		}
	} else {
		sync->load_log = fopen( fileName, "rb" );
	}

	if (sync->load_log == NULL) {
		perror("sync_LoadLog: fopen");
		return -1;
	}

#ifdef DEBUG_SYNC
	printf("Reading log from '%s'\n", fileName);
#endif

	sync->load_then_save = thenAppend;
	sync->simu_load = simLoad;

	return 0;
}

int sync_ConfigurationFinished( Synchronizer *sync )
{
int i;
int sock;

	if (sync->state != START_CONF) {
		printf("sync_ConfigurationFinished: illegal state.\n");
		return -1;
	}
	if (sync->num_out <= 0) {
		printf("sync_ConfigurationFinished: no out connections.\n");
		return -1;
	}

	sync->receive_counter = create_HashMap( &type_TE_ArcId, &type_Integer,
		DEFAULT_HASH_SIZE, (HashFunc)hash_TE_ArcId );

	sync->in_channels = sure_realloc( NULL, sync->num_in * sizeof(int) );
	sync->shutdown_received = sure_realloc( NULL, sync->num_in * sizeof(bool) );
	sync->read_poll = sure_realloc( NULL, sync->num_in * sizeof(int) );
	sync->in_failed = sure_realloc( NULL, sync->num_in * sizeof(bool) );

	sync->out_channels = sure_realloc( NULL, sync->num_out * sizeof(int) );
	sync->out_failed = sure_realloc( NULL, sync->num_out * sizeof(bool) );

	for (i = 0; i < sync->num_out; i++) {
		sock = connect_to_server( sync->out_infos[i].outHost, sync->out_infos[i].outPort );
		if (sock < 0)	return -1;
#ifdef DEBUG_SYNC
		printf("Out connection to %s:%i established.\n",
			sync->out_infos[i].outHost, (int)sync->out_infos[i].outPort);
#endif
		free( sync->out_infos[i].outHost );
		sync->out_channels[i] = sock;
		sync->out_failed[i] = false;
	}
	free( sync->out_infos );
	sync->out_infos = NULL;

	for (i = 0; i < sync->num_in; i++) {
		sock = accept_connection( sync->in_socket, true );
		if (sock < 0)	return -1;
#ifdef DEBUG_SYNC
		printf("In connection established.\n");
#endif
		sync->in_channels[i] = sock;
		sync->shutdown_received[i] = false;
		sync->in_failed[i] = false;
	}

	sync->state = READY;
	return 0;
}


/************************************************/
/*                  Threading                   */
/************************************************/

static void init_for_threads( Synchronizer *sync )
{
	sync->write_sema = ts_create_semaphore();
	sync->write_queue = create_List( &type_List );
}

void sync_SetExchangeDelay( Synchronizer *sync, int msecs )
{
	sync->thread_exchange_delay = msecs;
}

static void *reader_thread_main( void *s )
{
Synchronizer *sync = (Synchronizer *)s;

#ifdef DEBUG_SYNC
	printf("Synchronizer reader thread started.\n");
#endif

	while (!sync->thread_death_required && !sync->shutdown_activated) {
		TE_GraphStatus status = getStatus_Graph(sync->graph);
		if (status == TE_DESTROYED) {
#ifdef DEBUG_SYNC
			printf("Graph is already destroyed!\n");
#endif
			return NULL;
		}
		if (status == TE_STARTED) {
#ifdef DEBUG_SYNC
			printf("Graph have started.\n");
#endif
			break;
		}
		sleep_msec( sync->thread_exchange_delay );
	}

	if (sync->load_log != NULL)
		do_load_log( sync );

	while (!sync->thread_death_required && !sync->shutdown_activated) {
		if (! sync_DoExchange( sync ))
			sleep_msec( sync->thread_exchange_delay );
	}

#ifdef DEBUG_SYNC
	printf("Synchronizer reader thread finished.\n");
#endif
	return NULL;
}

static void *writer_thread_main( void *s )
{
Synchronizer *sync = (Synchronizer*)s;
List *message_lists = create_List( &type_List );
int i, j, len;

#ifdef DEBUG_SYNC
	printf("Synchronizer writer thread started.\n");
#endif

	while (!sync->thread_death_required && !sync->shutdown_activated) {
		ts_close_semaphore( sync->write_sema );
		if (sync->write_queue_size == 0) {
			ts_open_semaphore( sync->write_sema );
			sleep_msec( sync->thread_exchange_delay );
			continue;
		}

		len = size_List( r(sync->write_queue) );

#ifdef DEBUG_SYNC_HEAVY
		printf("Start draining outgoing queue...\n");
#endif
		appendAll_List( r(message_lists), r(sync->write_queue) );
		clear_List( r(sync->write_queue) );
		sync->writer_in_processing = 1;	/* To let someone lock until writer has finished */
		sync->write_queue_size = 0;
		ts_open_semaphore( sync->write_sema );

		for (i = 0; i < len; i++) {
			sync->writer_in_processing += size_List( get_List(r(message_lists),i) );
		}
		sync->writer_in_processing--; /* Compensate initial 1 */
#ifdef DEBUG_SYNC_HEAVY
		printf("Finished draining outgoing queue. %i message batches with total length %i arrived.\n",
			len, sync->writer_in_processing);
#endif

		if (sync->thread_death_required || sync->shutdown_activated)
			break;

		for (i = 0; i < len; i++) {
			List *list;
			int len2;
			list = get_List( r(message_lists), i );
			len2 = size_List( r(list) );
			for (j = 0; j < len2; j++) {
				if (sync->thread_death_required || sync->shutdown_activated)
					break;

				send_message_out( sync, get_List(r(list), j) );
				sync->writer_in_processing--;
			}
			destroy( list );

		}
		clear_List( r(message_lists) );
		sync->writer_in_processing = 0;
	}

	destroy( message_lists );
#ifdef DEBUG_SYNC
	printf("Synchronizer writer thread finished.\n");
#endif
	return NULL;
}


/*** sync_StartThread ***/
#if defined(WIN32)

static DWORD WINAPI ReaderThread(LPVOID lpParam)
{
	reader_thread_main( (Synchronizer*)lpParam );
	return 0;
}

static DWORD WINAPI WriterThread(LPVOID lpParam)
{
	writer_thread_main( (Synchronizer*)lpParam );
	return 0;
}

int sync_StartThread( Synchronizer *sync )
{
DWORD dwThreadId;

	if (sync->state != READY) {
		printf("sync_StartThread: illegal state.\n");
		return -1;
	}

	init_for_threads( sync );

	sync->state = THREAD_RUNNING;
	sync->writer_thread = CreateThread(NULL, 0, WriterThread, sync, 0, &dwThreadId);
	if (sync->writer_thread == 0) {
		printf("Create synchronizer writer thread failed: %d\n", GetLastError());
		sync->state = READY;
		return -1;
	}

	sync->reader_thread = CreateThread(NULL, 0, ReaderThread, sync, 0, &dwThreadId);
	if (sync->reader_thread == 0) {
		printf("Create synchronizer reader thread failed: %d\n", GetLastError());
		sync_FinishThread( sync );
		return -1;
	}
	return 0;
}

#elif defined(unix)

int sync_StartThread( Synchronizer *sync )
{
	if (sync->state != READY) {
		printf("sync_StartThread: illegal state.\n");
		return -1;
	}

	init_for_threads( sync );

	sync->state = THREAD_RUNNING;
    if (pthread_create( &sync->writer_thread, NULL, writer_thread_main, sync) != 0) {
		perror( "pthread_create: writer thread" );
		sync->state = READY;
        return -1;
    }

	sync->state = THREAD_RUNNING;
    if (pthread_create( &sync->reader_thread, NULL, reader_thread_main, sync) != 0) {
		perror( "pthread_create: reader thread" );
		sync_FinishThread( sync );
        return -1;
    }

	return 0;
}
#endif

/*** sync_FinishThread ***/
#if defined(WIN32)

static bool join_thread( HANDLE thread )
{
bool res;
	for (res = -1; res == -1; ) {
		DWORD r;
#ifdef DEBUG_SYNC
		printf("invoking WaitForSingleObject()\n");
		fflush(stdout);
#endif
		r = WaitForSingleObject( thread, INFINITE );
#ifdef DEBUG_SYNC
		printf("WaitForSingleObject() returned %li\n", (long int)r);
		fflush(stdout);
#endif
		switch (r)
		{
		case WAIT_OBJECT_0:
			res = true; break;
		case WAIT_FAILED:
			res = false; break;
		default:
			sleep_msec( 10);
		}
	}

	return CloseHandle( thread ) && res;
}

#elif defined(unix)

static bool join_thread( pthread_t thread )
{
	if (pthread_join( thread, NULL ) != 0) {
		perror( "pthread_join" );
		return false;
	}
	return true;
}
#endif

int sync_FinishThread( Synchronizer *sync )
{
bool success = true;

	if (sync->writer_thread == 0 && sync->reader_thread == 0) {
		printf("sync_FinishThread: illegal state.\n");
		return -1;
	}

#ifdef DEBUG_SYNC
	printf("sync_FinishThread(): the beginning of the end\n");
#endif

	sync->thread_death_required = true;

	if (sync->writer_thread == 0) {
		success = false;
	} else {
		success = join_thread( sync->writer_thread ) && success;
		sync->writer_thread = 0;
	}

	if (sync->reader_thread == 0) {
		success = false;
	} else {
		success = join_thread( sync->reader_thread ) && success;
		sync->reader_thread = 0;
	}

	/* Nobody writes to out channels, so we can safely send MTYPE_FINISH messages */
	sync_Shutdown( sync );	/* state --> FINALIZED */

	return success ? 0 : -1;
}

/************************************************/
/*                IO primitives                 */
/************************************************/

static void handle_read_error( Synchronizer *sync, int inChannel, int bytesExpected )
{
	if (inChannel == SRC_LOAD_LOG) {
		longjmp( sync->io_exception, 1 );
	}
	else if (sync->tolerate_neighbour_deaths) {
		printf( "[SYNCHRONIZER] Failure on input channel #%i\n", inChannel );
		sync->in_failed[ inChannel ] = true;
		sync->in_fail_count++;
		longjmp( sync->io_exception, 2 );
	}
	else ASSERT2( false, "Failed to read %i bytes from input channel #%i", bytesExpected, inChannel );
}

static int read_int( Synchronizer *sync, int inChannel )
{
int len, res;

	if (inChannel == SRC_LOAD_LOG) {
		len = fread( &res, 1, sizeof(int), sync->load_log );
	} else {
		int sock = sync->in_channels[ inChannel ];
		len = read_from_socket( sock, &res, sizeof(int) );
	}

	if (len != sizeof(int))
		handle_read_error( sync, inChannel, sizeof(int) );

	return res;
}

static String *read_string( Synchronizer *sync, int inChannel )
{
char small_buffer[ SMALL_MESSAGE_SIZE ];
char *s;
StringBuffer *sb;
int slen, len;
	
	slen = read_int( sync, inChannel );

	if (slen <= sizeof(small_buffer)) {
		s = small_buffer;
	} else {
		s = sure_realloc( NULL, slen);
	}

	if (inChannel == SRC_LOAD_LOG) {
		len = fread( s, 1, slen, sync->load_log );
	} else {
		int sock = sync->in_channels[ inChannel ];
		len = read_from_socket( sock, s, slen );
	}

	if (len != slen) {
		if (s != small_buffer)	free( s );

		handle_read_error( sync, inChannel, slen );
	}

	sb = createCustomized_StringBuffer( slen
		+ 1 /* for null byte added by to_string_StringBuffer() */ );
	appendCharArray_StringBuffer( r(sb), s, slen );

	if (s != small_buffer)	free( s );

	return toString(sb);
}

static TE_NodeId *read_state( Synchronizer *sync, int inChannel )
{
String *state_id = read_string( sync, inChannel );
int deg = read_int( sync, inChannel );
TE_NodeId *res;

	res = create_NodeId( state_id );
	res->deg = deg;
	return res;
}

static TE_ArcId *read_arc_message( Synchronizer *sync, int inChannel )
{
TE_NodeId *start_state, *end_state;
TE_StimulusId stimulus;
TE_ArcId *res;

	start_state = read_state( sync, inChannel );
	stimulus = create_StimulusId( read_int(sync, inChannel) );
	end_state = read_state( sync, inChannel );
	
	res = create_ArcId( start_state, stimulus, end_state );

	if (inChannel == SRC_LOAD_LOG)
		sync->loaded_arcs_count++;

#ifdef DEBUG_SYNC
	{
		String *ss = toString( r(res) );
		if (inChannel == SRC_LOAD_LOG)
			 printf("      Loaded: %s\n", toCharArray_String( r(ss) ));
		else printf("    Received: %s from input channel #%i\n", toCharArray_String( r(ss) ), inChannel);
		destroy(ss);
	}
#endif

	return res;
}

static String *serialize_arc( TE_ArcId *arc )
{
StringBuffer *sb;
int len1, len2, mtype;

	len1 = length_String( r(arc->start_node_id->id) );
	len2 = length_String( r(arc->end_node_id  ->id) );
	sb = createCustomized_StringBuffer( len1 + len2
		+ 6 * sizeof(int)	/* 1 mtype + 2 lengths + 2 degrees + 1 stimulus */
		+ 1 /* for null byte added by to_string_StringBuffer() */ );
	
	mtype = MTYPE_TRANS;
	appendCharArray_StringBuffer( r(sb), (char*)&mtype, sizeof(int) );

	appendCharArray_StringBuffer( r(sb), (char*)&len1, sizeof(int) );
	appendString_StringBuffer(    r(sb), r(arc->start_node_id->id) );
	appendCharArray_StringBuffer( r(sb), (char*)&arc->start_node_id->deg, sizeof(int) );
	
	appendCharArray_StringBuffer( r(sb), (char*)&arc->stimulus_id.index, sizeof(int) );

	appendCharArray_StringBuffer( r(sb), (char*)&len2, sizeof(int) );
	appendString_StringBuffer(    r(sb), r(arc->end_node_id->id) );
	appendCharArray_StringBuffer( r(sb), (char*)&arc->end_node_id->deg, sizeof(int) );

	destroy( arc );

	return toString(sb);
}

/************************************************/
/*                  Messaging                   */
/************************************************/

/** @return true if it was first addition */
static bool inc_arc_counter( Synchronizer *sync, TE_ArcId *arc )
{
Integer *count = get_HashMap( r(sync->receive_counter), r(arc) );
bool res;

#ifdef DEBUG_SYNC
String *ss = toString( r(arc) );
#endif

	if (count == NULL) {
		count = create_Integer(1);
		put_HashMap( r(sync->receive_counter), r(arc), r(count) ); /* silent prefix not needed */
		sync->receive_counter_size++;
		res = true;
	} else {
		(*count)++;	/* Violation(?) of Integer contract */
		res = false;
	}
#ifdef DEBUG_SYNC
	printf("+++Increment: %s, count = %i\n", toCharArray_String( r(ss) ), *count);
#endif

	if (*count == sync->num_in+1) {
#ifdef DEBUG_SYNC
		printf("++Forgetting: %s\n", toCharArray_String(r(ss)) );
#endif
		silentRemove_HashMap( r(sync->receive_counter), r(arc) );
		sync->receive_counter_size--;
		sync->completed_transactions++;
	}

	destroy(arc);
	destroy(count);
#ifdef DEBUG_SYNC
	destroy(ss);
#endif
	return res;
}

static void receive_message( Synchronizer *sync, int inChannel, /*no r()*/TE_ArcList *in_updates )
{
int mtype;
TE_ArcId *arc;

	mtype = read_int( sync, inChannel );

	switch (mtype) {
	case MTYPE_TRANS:
		arc = read_arc_message( sync, inChannel );
		if (in_updates != NULL)
			 append_List( r(in_updates), arc );
		else destroy(arc);
		break;

	case MTYPE_FINISH:
#ifdef DEBUG_SYNC
		printf("Shutdown message received from input channel #%i\n", inChannel);
#endif
		sync->shutdown_received[inChannel] = true;
		activate_shutdown( sync );
		break;

	default:
		ASSERT1( false, "Unknown message type: %i", mtype );
	}
}

static void send_message_out( Synchronizer *sync, TE_ArcId *arc )
{
String *msg = serialize_arc( r(arc) );
const char *s = toCharArray_String( r(msg) );
int bytes_sent, i, slen = length_String( r(msg) );

#ifdef DEBUG_SYNC
String *ss = toString( r(arc) );

	printf("     Sending: %s\n", toCharArray_String(r(ss)) );
	destroy(ss);
#endif

	if (sync->skip_save_log == 0 && sync->save_log != NULL) {
		bytes_sent = fwrite( s, 1, slen, sync->save_log );
		if (bytes_sent != slen) {
			printf( "Failed to write %i bytes to log. No more logging!!!\n", slen );
			fclose( sync->save_log );
			sync->save_log = NULL;
		} else {
			sync->saved_arcs_count++;
		}
	}
	if (sync->skip_save_log > 0) {
		sync->skip_save_log--;
#ifdef DEBUG_SYNC_HEAVY
		printf("Arc logging skipped, %d more left...\n", sync->skip_save_log);
#endif
	}

	for (i = 0; i < sync->num_out; i++) {
		if (sync->out_failed[i])	continue;

		bytes_sent = send_to_socket( sync->out_channels[i], s, slen );
		if (bytes_sent != slen) {
			if (sync->tolerate_neighbour_deaths) {
				printf( "[SYNCHRONIZER] Failure on output channel #%i\n", i );
				sync->out_failed[ i ] = true;
				sync->out_fail_count++;
			}
			else ASSERT2( false, "Failed to send %i bytes to output channel #%i", slen, i );
		}
	}

	destroy( msg );
	destroy( arc );
}


static void call_update_TE( Synchronizer *sync, List *update )
{
int len = size_List( r(update) );
int i;

	if (len <=  EXCHANGE_CHUNK) {
#ifdef DEBUG_SYNC_HEAVY
		printf("Calling updateRemoteAll_TestEngine() for %i arcs\n", len);
#endif
		updateRemoteAll_TestEngine( sync->te, sync->graph, /*no r()*/update );
	} else {
		for (i = 0; i < len; i += EXCHANGE_CHUNK) {
			List *subList;
			int upIndex = i + EXCHANGE_CHUNK;

			if (upIndex > len)	upIndex = len;
			subList = subList_List( r(update), i, upIndex );
#ifdef DEBUG_SYNC_HEAVY
			printf("Calling updateRemoteAll_TestEngine() for %i arcs\n", upIndex - i);
#endif
			updateRemoteAll_TestEngine( sync->te, sync->graph, /*no r()*/subList );
			destroy( subList );
		}
	}

	destroy( update );
}

static void enqueue_messages( Synchronizer *sync, List *messages, int len )
{
	ts_close_semaphore( sync->write_sema );
	sync->write_queue_size += len;
	append_List( r(sync->write_queue), messages );
	ts_open_semaphore( sync->write_sema );
}

static void send_update_part( Synchronizer *sync, List *update )
{
int len = size_List( r(update) );

#ifdef DEBUG_SYNC_HEAVY
	printf("Queuing batch of %i messages to writer thread\n", len);
#endif
	enqueue_messages( sync, update, len );
}

static void send_log_part( Synchronizer *sync, List *log )
{
int len = size_List( r(log) );

#ifdef DEBUG_SYNC_HEAVY
	String *ss = toString( getAllAttrValues_Synchronizer( sync ));
	printf( "%s\n", toCharArray_String(r(ss)) );
	destroy( ss );
	printf("Queuing loaded log contents for sending out, size = %i\n", len);
#endif

	enqueue_messages( sync, log, len );
}

/** @pre sync->load_log != NULL && called before any call to sync_DoExchange() */
static void do_load_log( Synchronizer *sync )
{
int i, len;
long success_position = 0;
List *log = createCustomized_List( &type_TE_ArcId, EXPECTED_LOG_SIZE );

/* <<<<<<<<<< Lock test engine */
	suspend_TestEngine( sync->te, sync->graph );

	if (!setjmp( sync->io_exception )) {
		while (!feof( sync->load_log ))	{
			/*read_arc_message(...)*/
			receive_message( sync, SRC_LOAD_LOG, /*no r()*/log );
			success_position = ftell( sync->load_log );
		}
	}
	/* else -- ignore errors in loading log */

	len = size_List( r(log) );

#ifdef DEBUG_SYNC
	printf ("Log file len = %ld, %ld bytes (%d arcs) loaded\n",
		ftell(sync->load_log), success_position, len);
#endif

	if (len > 0) {
		if (!sync->simu_load) {	/* Send loaded arcs to peers */
			List *log_copy = NULL;

			sync->previous_to_local = create_HashMap( &type_TE_ArcId, &type_String,
				len*5/3, (HashFunc)hash_TE_ArcId );

			if ( sync->load_then_save && sync->save_log == NULL ) {
				/* Log will be appended. Skip logging these arcs again. */
				sync->skip_save_log = len;
			}

			for (i = 0; i < len; i++) {
				TE_ArcId *arc = get_List( r(log), i );

/*				assertion( */

				inc_arc_counter( sync, r(arc) );

/*					"Synchronizer: load_log: log contains duplicate arc %s",
					toCharArray_String( r(toString(r(arc)))));
*/
				put_HashMap( r(sync->previous_to_local), r(arc), NULL );

				if (log_copy == NULL)
					log_copy = createCustomized_List( &type_TE_ArcId, EXCHANGE_CHUNK + 1 );

/*** Clone arc that will be enqueued for writer thread ***/
				append_List( r(log_copy), clone_Object( arc )); /* arc destroyed */
				if ( i > 0 && i % EXCHANGE_CHUNK == 0) {
					send_log_part( sync, log_copy );
					log_copy = NULL;
				}
			}
			if (log_copy != NULL)
				send_log_part( sync, log_copy );
		}
#ifdef DEBUG_SYNC
		else printf("Log contents is not sent to peers.\n");
#endif
	}

	resume_TestEngine( sync->te, sync->graph );
/* >>>>>>>>>> Unlock test engine */

	if (len > 0) {
#ifdef DEBUG_SYNC
		printf("Sending loaded log contents to Test Engine, size = %i\n", len);
#endif
/*** Clone arcs that will be added to local graph DB ***/
		call_update_TE( sync, clone_Object(r(log)) );
	}
	destroy( log );

	if ( sync->load_then_save && sync->save_log == NULL ) {
#ifdef DEBUG_SYNC
		printf ("Now appending to log at position %ld\n", success_position);
#endif
		sync->save_log = sync->load_log;
		fseek( sync->save_log, success_position, SEEK_SET );
		fwrite( &i, 1, 0, sync->save_log );
	} else {
		fclose( sync->load_log );
	}

	sync->load_log = NULL;
}

bool sync_DoExchange( Synchronizer *sync )
{
/* Note: contains_List(), add_Set() and contains_Set() take time
 proportional to the size of the target List/Set.
 We use HashMap to check for duplicates and List to preserve order,
 both working in singular time. */
TE_ArcList *in_updates, *local_updates, *out_updates, *to_local;
HashMap *hashed_to_local;

TE_ArcId *arc;
int i, j, len, in_len;

int pre_in_fail = sync->in_fail_count;

#ifdef DEBUG_SYNC
String *ss;
#endif

	if (sync->state != READY && sync->state != THREAD_RUNNING) {
		printf("sync_DoExchange: illegal state.\n");
		return false;
	}

#ifdef DEBUG_SYNC_HEAVY
	ss = toString( getAllAttrValues_Synchronizer( sync ));
	printf( "%s\n", toCharArray_String(r(ss)) );
	destroy( ss );
#endif

	/* Get incoming updates. Preserve all duplicates. */
	in_updates = createCustomized_List( &type_TE_ArcId, EXCHANGE_CHUNK );
	do {
		in_len = size_List( r(in_updates) );
		
		len = sync->num_in;
		memcpy( sync->read_poll, sync->in_channels, len * sizeof(int) );
		check_incoming_data( sync->read_poll, &len );

		j = 0;	/* Index of incoming connection */
		for (i = 0; i < len; i++) {
			while (sync->in_channels[j] != sync->read_poll[i])
				j++;
			
			if (!sync->in_failed[j] && !sync->shutdown_received[j]) {
				if (!sync->tolerate_neighbour_deaths || !setjmp( sync->io_exception )) {
					receive_message( sync, j, /*no r()*/in_updates );
				}
				/* else - everything is done in handle_read_error() */
			}

			j++;
		}
	} while (in_len != size_List(r(in_updates)));

	if (sync->shutdown_activated || sync->thread_death_required) {
		destroy( in_updates );
		return false;
	}

	/* Get local updates (always unique). Remove arcs added simultaneously
	 * by both local TE and synchronizer on previous exchange.
	 * Add new arcs to the outgoing queue. */
	local_updates = getUpdates_TestEngine( sync->te, sync->graph );
	len = size_List( r(local_updates) );
	
	if (in_len == 0 && len == 0) {
		destroy( in_updates );
		destroy( local_updates );
		return false;
	}

	out_updates = createCustomized_List( &type_TE_ArcId, in_len + len );
	to_local = createCustomized_List( &type_TE_ArcId, in_len );
	hashed_to_local = create_HashMap( &type_TE_ArcId, &type_String,
		in_len, (HashFunc)hash_TE_ArcId );

	for (i = 0; i < len /* == size(local_updates) */; i++) {
		if (sync->shutdown_activated || sync->thread_death_required)
			break;

		arc = get_List( r(local_updates), i );

		if (   sync->previous_to_local == NULL
			|| !containsKey_HashMap( r(sync->previous_to_local), r(arc) ))
		{
#ifdef DEBUG_SYNC
			ss = toString( r(arc) );
			printf("Local update: %s\n", toCharArray_String(r(ss)));
			destroy(ss);
#endif
			if (inc_arc_counter( sync, r(arc) ))
			{	/* Call to inc_arc_counter() must be guarded by check against
				previous_to_local. It triggers only once for every arc */
				append_List( r(out_updates), r(arc) );
			}
		}
#ifdef DEBUG_SYNC_HEAVY
		else {
			ss = toString( r(arc) );
			printf("RaceUpdateDB: %s\n", toCharArray_String( r(ss) ));
			destroy(ss);
		}
#endif

		destroy( arc );
	}


	/* Prepare incoming updates for retransmission and put them to local DB if needed */
	for (i = 0; i < in_len /* == size(in_updates) */; i++) {
		if (sync->shutdown_activated || sync->thread_death_required)
			break;

		arc = get_List( r(in_updates), i );

		if (inc_arc_counter( sync, r(arc) )) {	/* Will trigger only once for every arc */
#ifdef DEBUG_SYNC
			ss = toString( r(arc) );
			printf("Update -> DB: %s\n", toCharArray_String(r(ss)) );
			destroy(ss);
#endif
			append_List( r(out_updates), r(arc) );
/*** Clone arc that will be added to local graph DB ***/
			append_List( r(to_local), clone_Object(r(arc)) );
			put_HashMap( r(hashed_to_local), r(arc), NULL );
			inc_arc_counter( sync, r(arc) );	/* Extra increment: adding to local */
		}

		destroy(arc);
	}


	/* Update local DB with incoming arcs */
	len = size_List( r(to_local) );
	if (len > 0 && !sync->shutdown_activated && !sync->thread_death_required) {
#ifdef DEBUG_SYNC_HEAVY
		printf("Sending remote updates to Test Engine, size = %i\n", len);
#endif
		call_update_TE( sync, r(to_local) );
	}


	len = size_List( r(out_updates) );

	/* Treat new arcs as if they are already received from all old dead incoming connections.
	 New dead connections are unaffected, so some new arcs may hang in receive_counter forever. */
	if (pre_in_fail > 0) {
		for (i = 0; i < len; i++) {
			arc = get_List( r(out_updates), i );
#ifdef DEBUG_SYNC_HEAVY
			ss = toString( r(arc) );
			printf("DeadIncrement:%s %i times\n", toCharArray_String(r(ss)), pre_in_fail );
			destroy(ss);
#endif
			for (j = 0; j < pre_in_fail; j++)
				inc_arc_counter( sync, r(arc) );

			destroy( arc );
		}
	}

	/* Send messages out */
	if ( sync->write_queue != NULL ) {	/* Normal case (2 active threads). Deadlock-safe. */

		List *update_copy = NULL;
		for (i = 0; i < len; i++) {
			if (update_copy == NULL)
				update_copy = createCustomized_List( &type_TE_ArcId, EXCHANGE_CHUNK + 1 );
/*** Clone arc that will be enqueued for writer thread ***/
			append_List( r(update_copy), clone_Object( get_List(r(out_updates),i) ));

			if ( i > 0 && i % EXCHANGE_CHUNK == 0) {
				send_update_part( sync, update_copy );
				update_copy = NULL;
			}
		}

		if (update_copy != NULL)
			send_update_part( sync, update_copy );

	} else {	/* Passive synchronizer. May deadlock here. */
		for (i = 0; i < len; i++) {
			if (sync->shutdown_activated || sync->thread_death_required)
				break;

			send_message_out( sync, get_List( r(out_updates), i ) );
		}
	}


	/* Cleanup */
	destroy( sync->previous_to_local );
	sync->previous_to_local = hashed_to_local;

	destroy( in_updates );
	destroy( local_updates );
	destroy( out_updates );
	destroy( to_local );

	return true;
}

/************************************************/
/*               Shutdown protocol              */
/************************************************/

static void activate_shutdown( Synchronizer *sync )
{
	if (sync->shutdown_activated)
		return;

#ifdef DEBUG_SYNC
	printf("!!! Shutdown activated !!!\n");
#endif

	sync->shutdown_activated = true;
}

void sync_Shutdown( Synchronizer *sync )
{
int m, i;

	switch (sync->state) {
	case READY:
	case THREAD_RUNNING:
		break;
	case FINALIZED:
		return;
	default:
		printf("sync_Shutdown: illegal state.\n");
		return;
	}

#ifdef DEBUG_SYNC
	printf("!!! sync_Shutdown() called !!!\n");
#endif
	activate_shutdown( sync );

	for (i = 0; i < sync->num_out; i++) {
		if (!sync->out_failed[i]) {
#ifdef DEBUG_SYNC
			printf("Sending shutdown message to output channel #%i\n", i);
#endif
			m = MTYPE_FINISH;
			send_to_socket( sync->out_channels[i], &m, sizeof(int) );
		}
		close_socket( sync->out_channels[i] );
	}

	for (i = 0; i < sync->num_in; i++) {
		while (!sync->in_failed[i] && !sync->shutdown_received[i]) {
			receive_message( sync, i, NULL );
		}
		close_socket( sync->in_channels[i] );
	}

	close_socket(sync->in_socket);

	free( sync->in_channels );
	free( sync->shutdown_received );
	free( sync->read_poll );
	free( sync->out_channels );

	destroy( sync->previous_to_local );
	destroy( sync->receive_counter );

	if (sync->load_log != NULL)
		fclose( sync->load_log );
	if (sync->save_log != NULL)
		fclose( sync->save_log );

	if (sync->write_sema != NULL) {
		ts_destroy_semaphore( sync->write_sema );
		destroy( sync->write_queue );
	}

	sync->state = FINALIZED;
#ifdef DEBUG_SYNC
	printf("!!! sync_Shutdown() completed !!!\n");
#endif
}

/************************************************/
/*                    Stats                     */
/************************************************/

typedef String* (*AttrFunc)(Synchronizer *sync);

static char *get_status_str( enum Sync_State sync_state ) {
	switch (sync_state) {
		case CREATED:
			return SYNC_ATTR_STATUS_INIT;
		case START_CONF:
			return SYNC_ATTR_STATUS_SEMI_CONF;
		case READY:
			return SYNC_ATTR_STATUS_READY;
		case THREAD_RUNNING:
			return SYNC_ATTR_STATUS_RUNNING;
		case FINALIZED:
			return SYNC_ATTR_STATUS_FINALIZED;
		default:
			ASSERT0(false, "getAttrValue_Synchronizer: unknown status");
			return NULL;
	}
}

static String *get_status_String( Synchronizer *sync )
{
	if (sync->shutdown_activated)
		return create_String( SYNC_ATTR_STATUS_SHUTDOWN );
	return create_String( get_status_str(sync->state) );
}
static String *get_in_conns( Synchronizer *sync )
{
	return format_String("%i", sync->num_in);
}
static String *get_out_conns( Synchronizer *sync )
{
	return format_String("%i", sync->num_out);
}
static String *get_open_trans( Synchronizer *sync )
{
	return format_String("%i", sync->receive_counter_size );
}
static String *get_compl_trans( Synchronizer *sync )
{
	return format_String("%i", sync->completed_transactions );
}
static String *get_write_qlen( Synchronizer *sync )
{
	if (sync->write_sema != NULL) {
		return format_String("%i", sync->writer_in_processing + sync->write_queue_size);
	}
	else return NULL;
}
static String *get_loaded_arcs_count( Synchronizer *sync )
{
	return format_String("%i", sync->loaded_arcs_count);
}
static String *get_saved_arcs_count( Synchronizer *sync )
{
	return format_String("%i", sync->saved_arcs_count);
}
static String *get_logging_status( Synchronizer *sync )
{
	if (sync->load_log != NULL)
		return create_String("Loading");
	if (sync->save_log != NULL)
		return create_String("Saving");
	if (sync->saved_arcs_count > 0)
		return create_String("ABORTED");
	return create_String("None");
}
static String *get_in_failure_count( Synchronizer *sync )
{
	return format_String("%i", sync->in_fail_count);
}
static String *get_out_failure_count( Synchronizer *sync )
{
	return format_String("%i", sync->out_fail_count);
}

/************************************************/

static HashMap *attrFuncs;
static List *attrOrder;

static void addFunc( const char *attrName, AttrFunc func )
{
String *s = create_String(attrName);
	append_List( r(attrOrder), r(s) );
	put_HashMap( r(attrFuncs), s, create_VoidAst(func) );
}

static HashMap *getAttrFuncs() {
	if (attrFuncs != NULL)	return r(attrFuncs);

	attrFuncs = create_HashMap( &type_String, &type_VoidAst, 13, (HashFunc)hash_String );
	attrOrder = createCustomized_List( &type_String, 13 );

	addFunc( SYNC_ATTR_STATUS, get_status_String );
	addFunc( SYNC_ATTR_IN_CONNECTIONS, get_in_conns );
	addFunc( SYNC_ATTR_IN_FAILURES, get_in_failure_count );
	addFunc( SYNC_ATTR_OUT_CONNECTIONS, get_out_conns );
	addFunc( SYNC_ATTR_OUT_FAILURES, get_out_failure_count );
	addFunc( SYNC_ATTR_OPEN_TRANSACTIONS, get_open_trans );
	addFunc( SYNC_ATTR_COMPLETED_TRANSACTIONS, get_compl_trans );
	addFunc( SYNC_ATTR_OUT_QUEUE_SIZE, get_write_qlen );
	addFunc( SYNC_ATTR_LOGGING_STATUS, get_logging_status );
	addFunc( SYNC_ATTR_LOG_LOADED_ARCS, get_loaded_arcs_count );
	addFunc( SYNC_ATTR_LOG_SAVED_ARCS, get_saved_arcs_count );

	return r(attrFuncs);
}

static String* getAttrValue( Synchronizer *sync, String *attr_name )
{
AttrFunc func;
VoidAst *va = get_HashMap( getAttrFuncs(), attr_name );

	if (va == NULL)	return NULL;
	func = (AttrFunc)value_VoidAst( va );
	return (*func)(sync);
}

String* getAttrValue_Synchronizer( Synchronizer *sync, const char *attr_name )
{
	return getAttrValue( sync, create_String(attr_name) );
}

Map* getAllAttrValues_Synchronizer( Synchronizer *sync )
{
HashMap *hm = getAttrFuncs();
Map *map = create_Map( &type_String, &type_String );
int i, size;

	size = size_List( r(attrOrder) );
	for( i = 0; i < size; i++ ) {
		String *key = get_List( r(attrOrder), i );
		String *val = getAttrValue( sync, r(key) );
		if (val != NULL) {
			put_Map( r(map), r(key), r(val) );
		}
		destroy(key);
		destroy(val);
	}

	destroy( hm );
	return map;
}
